@@ -36,9 +36,8 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- *      %W% (Berkeley) %G%
  *
- * $Id: info_passwd.c,v 1.3.2.5 2005/01/03 20:56:12 ezk Exp $
+ * File: am-utils/amd/info_passwd.c
  *
  */
 
@@ -178,9 +177,7 @@ passwd_search(mnt_map *m, char *map, char *key, char **pval, time_t *tp)
       p = "type:=nfs;rfs:=/${var0}/${var1};rhost:=${var1};sublink:=${var2};fs:=${autodir}${var3}";
     sprintf(val, "var0:=%s;var1:=%s;var2:=%s;var3:=%s;%s",
 	    dir+1, rhost, user, pw->pw_dir, p);
-#ifdef DEBUG
     dlog("passwd_search: map=%s key=%s -> %s", map, key, val);
-#endif /* DEBUG */
     if (q)
       *q = '.';
     *pval = strdup(val);
