@@ -36,9 +36,8 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- *      %W% (Berkeley) %G%
  *
- * $Id: info_union.c,v 1.3.2.5 2005/01/03 20:56:12 ezk Exp $
+ * File: am-utils/amd/info_union.c
  *
  */
 
@@ -114,9 +113,7 @@ union_reload(mnt_map *m, char *map, void (*fn) (mnt_map *, char *, char *))
     }
     dlen = strlen(*dir);
 
-#ifdef DEBUG
     dlog("Reading directory %s...", *dir);
-#endif /* DEBUG */
     while ((dp = readdir(dirp))) {
       char *val, *dpname = &dp->d_name[0];
       if (dpname[0] == '.' &&
@@ -124,9 +121,7 @@ union_reload(mnt_map *m, char *map, void (*fn) (mnt_map *, char *, char *))
 	   (dpname[1] == '.' && dpname[2] == '\0')))
 	continue;
 
-#ifdef DEBUG
       dlog("... gives %s", dp->d_name);
-#endif /* DEBUG */
       val = xmalloc(dlen + 5);
       sprintf(val, "fs:=%s", *dir);
       (*fn) (m, strdup(dp->d_name), val);
@@ -138,9 +133,10 @@ union_reload(mnt_map *m, char *map, void (*fn) (mnt_map *, char *, char *))
    * Add wildcard entry
    */
   {
-    char *val = xmalloc(strlen(dir[-1]) + 5);
+    size_t len = strlen(*(dir-1)) + 5;
+    char *val = xmalloc(len);
 
-    sprintf(val, "fs:=%s", dir[-1]);
+    xsnprintf(val, len, "fs:=%s", *(dir-1));
     (*fn) (m, strdup("*"), val);
   }
   XFREE(mapd);
