@@ -36,9 +36,8 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- *      %W% (Berkeley) %G%
  *
- * $Id: mapc.c,v 1.7.2.9 2005/01/03 20:56:12 ezk Exp $
+ * File: am-utils/amd/mapc.c
  *
  */
 
@@ -85,8 +84,6 @@
 #define	MREC_PART	1
 #define	MREC_NONE	0
 
-#define MAX_CHAIN	2048
-
 static struct opt_tab mapc_opt[] =
 {
   {"all", MAPC_ALL},
@@ -201,12 +198,17 @@ extern int ndbm_search(mnt_map *, char *, char *, char **, time_t *);
 extern int ndbm_mtime(mnt_map *, char *, time_t *);
 #endif /* HAVE_MAP_NDBM */
 
+/* EXECUTABLE MAPS */
+#ifdef HAVE_MAP_EXEC
+extern int exec_init(mnt_map *, char *, time_t *);
+extern int exec_search(mnt_map *, char *, char *, char **, time_t *);
+#endif /* HAVE_MAP_EXEC */
+
 /* FILE MAPS */
 #ifdef HAVE_MAP_FILE
-extern int file_init(mnt_map *, char *, time_t *);
+extern int file_init_or_mtime(mnt_map *, char *, time_t *);
 extern int file_reload(mnt_map *, char *, add_fn *);
 extern int file_search(mnt_map *, char *, char *, char **, time_t *);
-extern int file_mtime(mnt_map *, char *, time_t *);
 #endif /* HAVE_MAP_FILE */
 
 
@@ -302,14 +304,25 @@ static map_type maptypes[] =
 #ifdef HAVE_MAP_FILE
   {
     "file",
-    file_init,
+    file_init_or_mtime,
     file_reload,
     NULL,			/* isup function */
     file_search,
-    file_mtime,
+    file_init_or_mtime,
     MAPC_ALL
   },
 #endif /* HAVE_MAP_FILE */
+#ifdef HAVE_MAP_EXEC
+  {
+    "exec",
+    exec_init,
+    error_reload,
+    NULL,			/* isup function */
+    exec_search,
+    error_mtime,
+    MAPC_INC
+  },
+#endif /* HAVE_MAP_EXEC */
   {
     "error",
     error_init,
@@ -339,14 +352,25 @@ kvhash_of(char *key)
 void
 mapc_showtypes(char *buf)
 {
-  map_type *mt;
-  char *sep = "";
+  map_type *mt=NULL, *lastmt;
+  int l = 0, i;
 
+  i = sizeof(maptypes) / sizeof(maptypes[0]);
+  lastmt = maptypes + i;
   buf[0] = '\0';
-  for (mt = maptypes; mt < maptypes + sizeof(maptypes) / sizeof(maptypes[0]); mt++) {
-    strcat(buf, sep);
+  for (mt = maptypes; mt < lastmt; mt++) {
     strcat(buf, mt->name);
-    sep = ", ";
+    if (mt == (lastmt-1))
+      break;	      /* if last one, don't do strcat's that follow */
+    l += strlen(mt->name);
+    if (--i > 0) {
+      strcat(buf, ", ");
+      l += 2;
+    }
+    if (l > 54) {
+      l = 0;
+      strcat(buf, "\n\t\t ");
+    }
   }
 }
 
@@ -386,9 +410,7 @@ mapc_add_kv(mnt_map *m, char *key, char *val)
   regex_t re;
 #endif /* HAVE_REGEXEC */
 
-#ifdef DEBUG
   dlog("add_kv: %s -> %s", key, val);
-#endif /* DEBUG */
 
 #ifdef HAVE_REGEXEC
   if (MAPC_ISRE(m)) {
@@ -512,10 +534,8 @@ mapc_reload_map(mnt_map *m)
   if (m->reloads != 0 && do_mapc_reload != 0) {
     if (t <= m->modify) {
       plog(XLOG_INFO, "reload of map %s is not needed (in sync)", m->map_name);
-#ifdef DEBUG
       dlog("map %s last load time is %d, last modify time is %d",
 	   m->map_name, (int) m->modify, (int) t);
-#endif /* DEBUG */
       return;
     }
   }
@@ -524,9 +544,7 @@ mapc_reload_map(mnt_map *m)
   memcpy((voidp) maphash, (voidp) m->kvhash, sizeof(m->kvhash));
   memset((voidp) m->kvhash, 0, sizeof(m->kvhash));
 
-#ifdef DEBUG
   dlog("calling map reload on %s", m->map_name);
-#endif /* DEBUG */
   error = (*m->reload) (m, m->map_name, mapc_add_kv);
   if (error) {
     if (m->reloads == 0)
@@ -550,9 +568,7 @@ mapc_reload_map(mnt_map *m)
   }
   m->wildcard = 0;
 
-#ifdef DEBUG
   dlog("calling mapc_search for wildcard");
-#endif /* DEBUG */
   error = mapc_search(m, wildcard, &m->wildcard);
   if (error)
     m->wildcard = 0;
@@ -603,9 +619,7 @@ mapc_create(char *map, char *opt, const char *type)
     for (mt = maptypes;
 	 mt < maptypes + sizeof(maptypes) / sizeof(maptypes[0]);
 	 mt++) {
-#ifdef DEBUG
       dlog("trying to initialize map %s of type %s ...", map, mt->name);
-#endif /* DEBUG */
       if ((*mt->init) (m, map, &modify) == 0) {
 	break;
       }
@@ -652,9 +666,7 @@ mapc_create(char *map, char *opt, const char *type)
 #endif /* HAVE_REGEXEC */
   }
 
-#ifdef DEBUG
   dlog("Map for %s coming from maptype %s", map, mt->name);
-#endif /* DEBUG */
 
   m->alloc = alloc;
   m->reload = mt->reload;
@@ -667,6 +679,8 @@ mapc_create(char *map, char *opt, const char *type)
   m->refc = 1;
   m->wildcard = 0;
   m->reloads = 0;
+  /* Unfortunately with current code structure, this cannot be initialized here */
+  m->cfm = NULL;
 
   /*
    * synchronize cache with reality
@@ -732,7 +746,7 @@ mapc_find(char *map, char *opt, const char *maptype)
    * add it to the list of maps
    */
   ITER(m, mnt_map, &map_list_head)
-  if (STREQ(m->map_name, map))
+    if (STREQ(m->map_name, map))
       return mapc_dup(m);
   m = mapc_create(map, opt, maptype);
   ins_que(&m->hdr, &map_list_head);
@@ -745,9 +759,9 @@ mapc_find(char *map, char *opt, const char *maptype)
  * Free a map.
  */
 void
-mapc_free(voidp v)
+mapc_free(opaque_t arg)
 {
-  mnt_map *m = v;
+  mnt_map *m = (mnt_map *) arg;
 
   /*
    * Decrement the reference count.
@@ -886,9 +900,7 @@ mapc_meta_search(mnt_map *m, char *key, char **pval, int recurse)
       strcpy(wildname, key);
       while (error && (subp = strrchr(wildname, '/'))) {
 	strcpy(subp, "/*");
-#ifdef DEBUG
 	dlog("mapc recurses on %s", wildname);
-#endif /* DEBUG */
 	error = mapc_meta_search(m, wildname, pval, MREC_PART);
 	if (error)
 	  *subp = 0;
@@ -1007,7 +1019,8 @@ root_newmap(const char *dir, const char *opts, const char *map, const cf_map_t *
 
   if (cfm) {
     if (map) {
-      sprintf(str, "cache:=mapdefault;type:=toplvl;fs:=\"%s\"",
+      sprintf(str, "cache:=mapdefault;type:=toplvl;mount_type:=%s;fs:=\"%s\"",
+	      cfm->cfm_flags & CFM_MOUNT_TYPE_AUTOFS ? "autofs" : "nfs",
 	      get_full_path(map, cfm->cfm_search_path, cfm->cfm_type));
       if (opts && opts[0] != '\0') {
 	strcat(str, ";");
@@ -1036,7 +1049,7 @@ root_newmap(const char *dir, const char *opts, const char *map, const cf_map_t *
 
 
 int
-mapc_keyiter(mnt_map *m, void (*fn) (char *, voidp), voidp arg)
+mapc_keyiter(mnt_map *m, key_fun *fn, opaque_t arg)
 {
   int i;
   int c = 0;
@@ -1056,10 +1069,10 @@ mapc_keyiter(mnt_map *m, void (*fn) (char *, voidp), voidp arg)
 
 /*
  * Iterate on the root map and call (*fn)() on the key of all the nodes.
- * Finally throw away the root map.
+ * Returns the number of entries in the root map.
  */
 int
-root_keyiter(void (*fn)(char *, voidp), voidp arg)
+root_keyiter(key_fun *fn, opaque_t arg)
 {
   if (root_map) {
     int c = mapc_keyiter(root_map, fn, arg);
@@ -1070,133 +1083,6 @@ root_keyiter(void (*fn)(char *, voidp), voidp arg)
 }
 
 
-/*
- * Was: NEW_TOPLVL_READDIR
- * Search a chain for an entry with some name.
- * -Erez Zadok <ezk@cs.columbia.edu>
- */
-static int
-key_already_in_chain(char *keyname, const nfsentry *chain)
-{
-  const nfsentry *tmpchain = chain;
-
-  while (tmpchain) {
-    if (keyname && tmpchain->ne_name && STREQ(keyname, tmpchain->ne_name))
-        return 1;
-    tmpchain = tmpchain->ne_nextentry;
-  }
-
-  return 0;
-}
-
-
-/*
- * Create a chain of entries which are not linked.
- * -Erez Zadok <ezk@cs.columbia.edu>
- */
-nfsentry *
-make_entry_chain(am_node *mp, const nfsentry *current_chain, int fully_browsable)
-{
-  static u_int last_cookie = (u_int) 2;	/* monotonically increasing */
-  static nfsentry chain[MAX_CHAIN];
-  static int max_entries = MAX_CHAIN;
-  char *key;
-  int num_entries = 0, i;
-  u_int preflen = 0;
-  nfsentry *retval = (nfsentry *) NULL;
-  mntfs *mf;
-  mnt_map *mmp;
-
-  if (!mp) {
-    plog(XLOG_DEBUG, "make_entry_chain: mp is (NULL)");
-    return retval;
-  }
-  mf = mp->am_mnt;
-  if (!mf) {
-    plog(XLOG_DEBUG, "make_entry_chain: mp->am_mnt is (NULL)");
-    return retval;
-  }
-  mmp = (mnt_map *) mf->mf_private;
-  if (!mmp) {
-    plog(XLOG_DEBUG, "make_entry_chain: mp->am_mnt->mf_private is (NULL)");
-    return retval;
-  }
-
-  if (mp->am_pref)
-    preflen = strlen(mp->am_pref);
-
-  /* iterate over keys */
-  for (i = 0; i < NKVHASH; i++) {
-    kv *k;
-    for (k = mmp->kvhash[i]; k ; k = k->next) {
-
-      /*
-       * Skip unwanted entries which are either not real entries or
-       * very difficult to interpret (wildcards...)  This test needs
-       * lots of improvement.  Any takers?
-       */
-      key = k->key;
-      if (!key)
-	continue;
-
-      /* Skip '*' */
-      if (!fully_browsable && strchr(key, '*'))
-	continue;
-
-      /*
-       * If the map has a prefix-string then check if the key starts with
-       * this string, and if it does, skip over this prefix.  If it has a
-       * prefix and it doesn't match the start of the key, skip it.
-       */
-      if (preflen && (preflen <= (strlen(key)))) {
-	if (!NSTREQ(key, mp->am_pref, preflen))
-	  continue;
-	key += preflen;
-      } else if (preflen) {
-	  continue;
-      }
-
-      /* no more '/' are allowed, unless browsable_dirs=full was used */
-      if (!fully_browsable && strchr(key, '/'))
-	continue;
-
-      /* no duplicates allowed */
-      if (key_already_in_chain(key, current_chain))
-	continue;
-
-      /* fill in a cell and link the entry */
-      if (num_entries >= max_entries) {
-	/* out of space */
-	plog(XLOG_DEBUG, "make_entry_chain: no more space in chain");
-	if (num_entries > 0) {
-	  chain[num_entries - 1].ne_nextentry = 0;
-	  retval = &chain[0];
-	}
-	return retval;
-      }
-
-      /* we have space.  put entry in next cell */
-      ++last_cookie;
-      chain[num_entries].ne_fileid = (u_int) last_cookie;
-      *(u_int *) chain[num_entries].ne_cookie = (u_int) last_cookie;
-      chain[num_entries].ne_name = key;
-      if (num_entries < max_entries - 1) {	/* link to next one */
-	chain[num_entries].ne_nextentry = &chain[num_entries + 1];
-      }
-      ++num_entries;
-    } /* end of "while (k)" */
-  } /* end of "for (i ... NKVHASH ..." */
-
-  /* terminate chain */
-  if (num_entries > 0) {
-    chain[num_entries - 1].ne_nextentry = 0;
-    retval = &chain[0];
-  }
-
-  return retval;
-}
-
-
 /*
  * Error map
  */
