@@ -36,9 +36,8 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- *      %W% (Berkeley) %G%
  *
- * $Id: transp_tli.c,v 1.5.2.8 2005/01/03 20:56:13 ezk Exp $
+ * File: am-utils/conf/transp/transp_tli.c
  *
  * TLI specific utilities.
  *      -Erez Zadok <ezk@cs.columbia.edu>
@@ -52,12 +51,16 @@
 
 struct netconfig *nfsncp;
 
+/* provide a definition for systems that don't have this */
+#ifndef INADDR_LOOPBACK
+# define INADDR_LOOPBACK	0x7f000001
+#endif /* not INADDR_LOOPBACK */
 
 /*
  * find the IP address that can be used to connect to the local host
  */
 void
-amu_get_myaddress(struct in_addr *iap)
+amu_get_myaddress(struct in_addr *iap, const char *preferred_localhost)
 {
   int ret;
   voidp handlep;
@@ -67,21 +70,25 @@ amu_get_myaddress(struct in_addr *iap)
 
   handlep = setnetconfig();
   ncp = getnetconfig(handlep);
-  service.h_host = HOST_SELF_CONNECT;
+  service.h_host = (preferred_localhost ? (char *) preferred_localhost : HOST_SELF_CONNECT);
   service.h_serv = (char *) NULL;
 
   ret = netdir_getbyname(ncp, &service, &addrs);
 
   if (ret || !addrs || addrs->n_cnt < 1) {
     plog(XLOG_FATAL, "cannot get local host address. using 127.0.0.1");
-    iap->s_addr = 0x7f000001;
+    iap->s_addr = htonl(INADDR_LOOPBACK);
   } else {
     /*
      * XXX: there may be more more than one address for this local
      * host.  Maybe something can be done with those.
      */
     struct sockaddr_in *sinp = (struct sockaddr_in *) addrs->n_addrs[0].buf;
-    iap->s_addr = htonl(sinp->sin_addr.s_addr);
+    char dq[20];
+    if (preferred_localhost)
+      plog(XLOG_INFO, "localhost_address \"%s\" requested, using %s",
+	   preferred_localhost, inet_dquad(dq, iap->s_addr));
+    iap->s_addr = sinp->sin_addr.s_addr; /* XXX: used to be htonl() */
   }
 
   endnetconfig(handlep);	/* free's up internal resources too */
@@ -114,7 +121,7 @@ bind_resv_port(int td, u_short *pp)
   memset((char *) treq->addr.buf, 0, treq->addr.len);
   sin = (struct sockaddr_in *) treq->addr.buf;
   sin->sin_family = AF_INET;
-  treq->qlen = 0;
+  treq->qlen = 64; /* 0 is ok for udp, for tcp you need qlen>0 */
   treq->addr.len = treq->addr.maxlen;
   errno = EADDRINUSE;
   port = IPPORT_RESERVED;
@@ -145,94 +152,6 @@ bind_resv_port(int td, u_short *pp)
 }
 
 
-/*
- * How to bind to reserved ports.
- * (port-only) version.
- */
-int
-bind_resv_port2(u_short *pp)
-{
-  int td, rc = -1, port;
-  struct t_bind *treq, *tret;
-  struct sockaddr_in *sin;
-  extern char *t_errlist[];
-  extern int t_errno;
-  struct netconfig *nc = (struct netconfig *) NULL;
-  voidp nc_handle;
-
-  if ((nc_handle = setnetconfig()) == (voidp) NULL) {
-    plog(XLOG_ERROR, "Cannot rewind netconfig: %s", nc_sperror());
-    return -1;
-  }
-  /*
-   * Search the netconfig table for INET/UDP.
-   * This loop will terminate if there was an error in the /etc/netconfig
-   * file or if you reached the end of the file without finding the udp
-   * device.  Either way your machine has probably far more problems (for
-   * example, you cannot have nfs v2 w/o UDP).
-   */
-  while (1) {
-    if ((nc = getnetconfig(nc_handle)) == (struct netconfig *) NULL) {
-      plog(XLOG_ERROR, "Error accessing getnetconfig: %s", nc_sperror());
-      endnetconfig(nc_handle);
-      return -1;
-    }
-    if (STREQ(nc->nc_protofmly, NC_INET) &&
-	STREQ(nc->nc_proto, NC_UDP))
-      break;
-  }
-
-  /*
-   * This is the primary reason for the getnetconfig code above: to get the
-   * correct device name to udp, and t_open a descriptor to be used in
-   * t_bind below.
-   */
-  td = t_open(nc->nc_device, O_RDWR, (struct t_info *) 0);
-  endnetconfig(nc_handle);
-
-  if (td < 0) {
-    plog(XLOG_ERROR, "t_open failed: %d: %s", t_errno, t_errlist[t_errno]);
-    return -1;
-  }
-  treq = (struct t_bind *) t_alloc(td, T_BIND, T_ADDR);
-  if (!treq) {
-    plog(XLOG_ERROR, "t_alloc req");
-    return -1;
-  }
-  tret = (struct t_bind *) t_alloc(td, T_BIND, T_ADDR);
-  if (!tret) {
-    t_free((char *) treq, T_BIND);
-    plog(XLOG_ERROR, "t_alloc ret");
-    return -1;
-  }
-  memset((char *) treq->addr.buf, 0, treq->addr.len);
-  sin = (struct sockaddr_in *) treq->addr.buf;
-  sin->sin_family = AF_INET;
-  treq->qlen = 0;
-  treq->addr.len = treq->addr.maxlen;
-  errno = EADDRINUSE;
-  port = IPPORT_RESERVED;
-
-  do {
-    --port;
-    sin->sin_port = htons(port);
-    rc = t_bind(td, treq, tret);
-    if (rc < 0) {
-      plog(XLOG_ERROR, "t_bind for port %d: %s", port, t_errlist[t_errno]);
-    } else {
-      if (memcmp(treq->addr.buf, tret->addr.buf, tret->addr.len) == 0)
-	break;
-      else
-	t_unbind(td);
-    }
-  } while ((rc < 0 || errno == EADDRINUSE) && (int) port > IPPORT_RESERVED / 2);
-
-  if (pp && rc == 0)
-    *pp = port;
-  t_free((char *) tret, T_BIND);
-  t_free((char *) treq, T_BIND);
-  return rc;
-}
 
 
 /*
@@ -296,9 +215,7 @@ get_mount_client(char *host, struct sockaddr_in *unused_sin, struct timeval *tv,
     goto tryudp;
   }
   /* tcp succeeded */
-#ifdef DEBUG
   dlog("get_mount_client: using tcp, port %d", sin.sin_port);
-#endif /* DEBUG */
   if (nc)
     freenetconfigent(nc);
   return client;
@@ -348,9 +265,7 @@ tryudp:
     goto badout;		/* neither tcp not udp succeeded */
   }
   /* udp succeeded */
-#ifdef DEBUG
   dlog("get_mount_client: using udp, port %d", sin.sin_port);
-#endif /* DEBUG */
   return client;
 
 badout:
@@ -361,18 +276,129 @@ badout:
 }
 
 
+#ifdef NOT_NEEDED_ON_TLI_SYSTEMS
 /*
  * find the address of the caller of an RPC procedure.
  */
 struct sockaddr_in *
 amu_svc_getcaller(SVCXPRT *xprt)
 {
-  struct netbuf *nbp = (struct netbuf *) NULL;
+  /*
+   * On TLI systems we don't use an INET network type, but a "ticlts" (see
+   * /etc/netconfig).  This means that packets could only come from the
+   * loopback interface, and we don't need to check them and filter possibly
+   * spoofed packets.  Therefore we simply return NULL here, and the caller
+   * will ignore the result.
+   */
+  return NULL;			/* tell called to ignore check */
+}
+#endif /* NOT_NEEDED_ON_TLI_SYSTEMS */
 
-  if ((nbp = svc_getrpccaller(xprt)) != NULL)
-    return (struct sockaddr_in *) nbp->buf; /* all OK */
 
-  return NULL;			/* failed */
+/*
+ * register an RPC server
+ */
+int
+amu_svc_register(SVCXPRT *xprt, u_long prognum, u_long versnum, void (*dispatch)(), u_long protocol, struct netconfig *ncp)
+{
+  return svc_reg(xprt, prognum, versnum, dispatch, ncp);
+}
+
+
+/*
+ * Bind to reserved UDP port, for NFS service only.
+ * Port-only version.
+ */
+static int
+bind_resv_port_only_udp(u_short *pp)
+{
+  int td, rc = -1, port;
+  struct t_bind *treq, *tret;
+  struct sockaddr_in *sin;
+  extern char *t_errlist[];
+  extern int t_errno;
+  struct netconfig *nc = (struct netconfig *) NULL;
+  voidp nc_handle;
+
+  if ((nc_handle = setnetconfig()) == (voidp) NULL) {
+    plog(XLOG_ERROR, "Cannot rewind netconfig: %s", nc_sperror());
+    return -1;
+  }
+  /*
+   * Search the netconfig table for INET/UDP.
+   * This loop will terminate if there was an error in the /etc/netconfig
+   * file or if you reached the end of the file without finding the udp
+   * device.  Either way your machine has probably far more problems (for
+   * example, you cannot have nfs v2 w/o UDP).
+   */
+  while (1) {
+    if ((nc = getnetconfig(nc_handle)) == (struct netconfig *) NULL) {
+      plog(XLOG_ERROR, "Error accessing getnetconfig: %s", nc_sperror());
+      endnetconfig(nc_handle);
+      return -1;
+    }
+    if (STREQ(nc->nc_protofmly, NC_INET) &&
+	STREQ(nc->nc_proto, NC_UDP))
+      break;
+  }
+
+  /*
+   * This is the primary reason for the getnetconfig code above: to get the
+   * correct device name to udp, and t_open a descriptor to be used in
+   * t_bind below.
+   */
+  td = t_open(nc->nc_device, O_RDWR, (struct t_info *) 0);
+  endnetconfig(nc_handle);
+
+  if (td < 0) {
+    plog(XLOG_ERROR, "t_open failed: %d: %s", t_errno, t_errlist[t_errno]);
+    return -1;
+  }
+  treq = (struct t_bind *) t_alloc(td, T_BIND, T_ADDR);
+  if (!treq) {
+    plog(XLOG_ERROR, "t_alloc req");
+    return -1;
+  }
+  tret = (struct t_bind *) t_alloc(td, T_BIND, T_ADDR);
+  if (!tret) {
+    t_free((char *) treq, T_BIND);
+    plog(XLOG_ERROR, "t_alloc ret");
+    return -1;
+  }
+  memset((char *) treq->addr.buf, 0, treq->addr.len);
+  sin = (struct sockaddr_in *) treq->addr.buf;
+  sin->sin_family = AF_INET;
+  treq->qlen = 64; /* 0 is ok for udp, for tcp you need qlen>0 */
+  treq->addr.len = treq->addr.maxlen;
+  errno = EADDRINUSE;
+
+  if (pp && *pp > 0) {
+    sin->sin_port = htons(*pp);
+    rc = t_bind(td, treq, tret);
+  } else {
+    port = IPPORT_RESERVED;
+
+    do {
+      --port;
+      sin->sin_port = htons(port);
+      rc = t_bind(td, treq, tret);
+      if (rc < 0) {
+	plog(XLOG_ERROR, "t_bind for port %d: %s", port, t_errlist[t_errno]);
+      } else {
+	if (memcmp(treq->addr.buf, tret->addr.buf, tret->addr.len) == 0)
+	  break;
+	else
+	  t_unbind(td);
+      }
+    } while ((rc < 0 || errno == EADDRINUSE) && (int) port > IPPORT_RESERVED / 2);
+
+    if (pp && rc == 0)
+      *pp = port;
+  }
+
+  t_free((char *) tret, T_BIND);
+  t_free((char *) treq, T_BIND);
+  return rc;
 }
 
 
@@ -380,10 +406,10 @@ amu_svc_getcaller(SVCXPRT *xprt)
  * Bind NFS to a reserved port.
  */
 static int
-bindnfs_port(int unused_so, u_short *nfs_portp)
+bind_nfs_port(int unused_so, u_short *nfs_portp)
 {
-  u_short port;
-  int error = bind_resv_port2(&port);
+  u_short port = 0;
+  int error = bind_resv_port_only_udp(&port);
 
   if (error == 0)
     *nfs_portp = port;
@@ -423,12 +449,14 @@ create_nfs_service(int *soNFSp, u_short *nfs_portp, SVCXPRT **nfs_xprtp, void (*
    * reserved nfs port.
    */
   *soNFSp = (*nfs_xprtp)->xp_fd;
-  if (*soNFSp < 0 || bindnfs_port(*soNFSp, nfs_portp) < 0) {
+  if (*soNFSp < 0 || bind_nfs_port(*soNFSp, nfs_portp) < 0) {
     plog(XLOG_ERROR, "Can't create privileged nfs port (TLI)");
+    svc_destroy(*nfs_xprtp);
     return 1;
   }
   if (svc_reg(*nfs_xprtp, NFS_PROGRAM, NFS_VERSION, dispatch_fxn, NULL) != 1) {
     plog(XLOG_ERROR, "could not register amd NFS service");
+    svc_destroy(*nfs_xprtp);
     return 1;
   }
 
@@ -436,15 +464,94 @@ create_nfs_service(int *soNFSp, u_short *nfs_portp, SVCXPRT **nfs_xprtp, void (*
 }
 
 
+/*
+ * Bind to preferred AMQ port.
+ */
+static int
+bind_preferred_amq_port(u_short pref_port,
+			const struct netconfig *ncp,
+			struct t_bind **tretpp)
+{
+  int td = -1, rc = -1;
+  struct t_bind *treq;
+  struct sockaddr_in *sin, *sin2;
+  extern char *t_errlist[];
+  extern int t_errno;
+
+  if (!ncp) {
+    plog(XLOG_ERROR, "null ncp");
+    return -1;
+  }
+
+  td = t_open(ncp->nc_device, O_RDWR, (struct t_info *) 0);
+  if (td < 0) {
+    plog(XLOG_ERROR, "t_open failed: %d: %s", t_errno, t_errlist[t_errno]);
+    return -1;
+  }
+  treq = (struct t_bind *) t_alloc(td, T_BIND, T_ADDR);
+  if (!treq) {
+    plog(XLOG_ERROR, "t_alloc req");
+    return -1;
+  }
+  *tretpp = (struct t_bind *) t_alloc(td, T_BIND, T_ADDR);
+  if (!*tretpp) {
+    t_free((char *) treq, T_BIND);
+    plog(XLOG_ERROR, "t_alloc tretpp");
+    return -1;
+  }
+  memset((char *) treq->addr.buf, 0, treq->addr.len);
+  sin = (struct sockaddr_in *) treq->addr.buf;
+  sin->sin_family = AF_INET;
+  treq->qlen = 64; /* must be greater than 0 to work for TCP connections */
+  treq->addr.len = treq->addr.maxlen;
+
+  if (pref_port > 0) {
+    sin->sin_port = htons(pref_port);
+    sin->sin_addr.s_addr = htonl(INADDR_LOOPBACK); /* XXX: may not be needed */
+    rc = t_bind(td, treq, *tretpp);
+    if (rc < 0) {
+      plog(XLOG_ERROR, "t_bind return err %d", rc);
+      goto out;
+    }
+    /* check if we got the port we asked for */
+    sin2 = (struct sockaddr_in *) (*tretpp)->addr.buf;
+    if (sin->sin_port != sin2->sin_port) {
+      plog(XLOG_ERROR, "asked for port %d, got different one (%d)",
+	   ntohs(sin->sin_port), ntohs(sin2->sin_port));
+      t_errno = TNOADDR; /* XXX: is this correct? */
+      rc = -1;
+      goto out;
+    }
+    if (sin->sin_addr.s_addr != sin2->sin_addr.s_addr) {
+      plog(XLOG_ERROR, "asked for address %x, got different one (%x)",
+	   (int) ntohl(sin->sin_addr.s_addr), (int) ntohl(sin2->sin_addr.s_addr));
+      t_errno = TNOADDR; /* XXX: is this correct? */
+      rc = -1;
+      goto out;
+    }
+  }
+out:
+  t_free((char *) treq, T_BIND);
+  return (rc < 0 ? rc : td);
+}
+
+
 /*
  * Create the amq service for amd (both TCP and UDP)
  */
 int
-create_amq_service(int *udp_soAMQp, SVCXPRT **udp_amqpp, struct netconfig **udp_amqncpp, int *tcp_soAMQp, SVCXPRT **tcp_amqpp, struct netconfig **tcp_amqncpp)
+create_amq_service(int *udp_soAMQp,
+		   SVCXPRT **udp_amqpp,
+		   struct netconfig **udp_amqncpp,
+		   int *tcp_soAMQp,
+		   SVCXPRT **tcp_amqpp,
+		   struct netconfig **tcp_amqncpp,
+		   u_short preferred_amq_port)
 {
   /*
    * (partially) create the amq service for amd
    * to be completed further in by caller.
+   * XXX: is this "partially" still true?!  See amd/nfs_start.c. -Erez
    */
 
   /* first create the TCP service */
@@ -453,8 +560,27 @@ create_amq_service(int *udp_soAMQp, SVCXPRT **udp_amqpp, struct netconfig **udp_
       plog(XLOG_ERROR, "cannot getnetconfigent for %s", NC_TCP);
       return 1;
     }
+
   if (tcp_amqpp) {
-    *tcp_amqpp = svc_tli_create(RPC_ANYFD, *tcp_amqncpp, NULL, 0, 0);
+    if (preferred_amq_port > 0) {
+      struct t_bind *tbp = NULL;
+      int sock;
+
+      plog(XLOG_INFO, "requesting preferred amq TCP port %d", preferred_amq_port);
+      sock = bind_preferred_amq_port(preferred_amq_port, *tcp_amqncpp, &tbp);
+      if (sock < 0) {
+	plog(XLOG_ERROR, "bind_preferred_amq_port failed for TCP port %d: %s",
+	     preferred_amq_port, t_errlist[t_errno]);
+	return 1;
+      }
+      *tcp_amqpp = svc_tli_create(sock, *tcp_amqncpp, tbp, 0, 0);
+      if (*tcp_amqpp != NULL)
+	plog(XLOG_INFO, "amq service bound to TCP port %d", preferred_amq_port);
+      t_free((char *) tbp, T_BIND);
+    } else {
+      /* select any port */
+      *tcp_amqpp = svc_tli_create(RPC_ANYFD, *tcp_amqncpp, NULL, 0, 0);
+    }
     if (*tcp_amqpp == NULL) {
       plog(XLOG_ERROR, "cannot create (tcp) tli service for amq");
       return 1;
@@ -470,7 +596,25 @@ create_amq_service(int *udp_soAMQp, SVCXPRT **udp_amqpp, struct netconfig **udp_
       return 1;
     }
   if (udp_amqpp) {
-    *udp_amqpp = svc_tli_create(RPC_ANYFD, *udp_amqncpp, NULL, 0, 0);
+    if (preferred_amq_port > 0) {
+      struct t_bind *tbp = NULL;
+      int sock;
+
+      plog(XLOG_INFO, "requesting preferred amq UDP port %d", preferred_amq_port);
+      sock = bind_preferred_amq_port(preferred_amq_port, *udp_amqncpp, &tbp);
+      if (sock < 0) {
+	plog(XLOG_ERROR, "bind_preferred_amq_port failed for UDP port %d: %s",
+	     preferred_amq_port, t_errlist[t_errno]);
+	return 1;
+      }
+      *udp_amqpp = svc_tli_create(sock, *udp_amqncpp, tbp, 0, 0);
+      if (*udp_amqpp != NULL)
+	plog(XLOG_INFO, "amq service bound to UDP port %d", preferred_amq_port);
+      t_free((char *) tbp, T_BIND);
+    } else {
+      /* select any port */
+      *udp_amqpp = svc_tli_create(RPC_ANYFD, *udp_amqncpp, NULL, 0, 0);
+    }
     if (*udp_amqpp == NULL) {
       plog(XLOG_ERROR, "cannot create (udp) tli service for amq");
       return 1;
@@ -551,82 +695,47 @@ free_knetconfig(struct knetconfig *kncp)
 }
 
 
-/* get the best possible NFS version for a host and transport */
-static CLIENT *
-amu_clnt_create_best_vers(const char *hostname, u_long program, u_long *out_version, u_long low_version, u_long high_version, const char *nettype)
+/*
+ * Check if the portmapper is running and reachable: 0==down, 1==up
+ */
+int check_pmap_up(char *host, struct sockaddr_in* sin)
 {
-  CLIENT *clnt;
-  enum clnt_stat rpc_stat;
-  struct rpc_err rpcerr;
-  struct timeval tv;
-  u_long lo, hi;
+  CLIENT *client;
+  enum clnt_stat clnt_stat = RPC_TIMEDOUT; /* assume failure */
+  int socket = RPC_ANYSOCK;
+  struct timeval timeout;
 
-  /* 3 seconds is more than enough for a LAN */
-  tv.tv_sec = 3;
-  tv.tv_usec = 0;
-
-#ifdef HAVE_CLNT_CREATE_TIMED
-  clnt = clnt_create_timed(hostname, program, high_version, nettype, &tv);
-  if (!clnt) {
-    plog(XLOG_INFO, "failed to create RPC client to \"%s\" after %d seconds",
-	 hostname, (int) tv.tv_sec);
-    return NULL;
+  timeout.tv_sec = 2;
+  timeout.tv_usec = 0;
+  sin->sin_port = htons(PMAPPORT);
+  client = clntudp_create(sin, PMAPPROG, PMAPVERS, timeout, &socket);
+  if (client == (CLIENT *) NULL) {
+    plog(XLOG_ERROR,
+	 "check_pmap_up: cannot create connection to contact portmapper on host \"%s\"%s",
+	 host, clnt_spcreateerror(""));
+    return 0;
   }
-#else /* not HAVE_CLNT_CREATE_TIMED */
-  /* Solaris 2.3 and earlier didn't have clnt_create_timed() */
-  clnt = clnt_create(hostname, program, high_version, nettype);
-  if (!clnt) {
-    plog(XLOG_INFO, "failed to create RPC client to \"%s\"", hostname);
-    return NULL;
-  }
-#endif /* not HAVE_CLNT_CREATE_TIMED */
 
-  rpc_stat = clnt_call(clnt,
-		       NULLPROC,
-		       (XDRPROC_T_TYPE) xdr_void,
-		       NULL,
-		       (XDRPROC_T_TYPE) xdr_void,
-		       NULL,
-		       tv);
-  if (rpc_stat == RPC_SUCCESS) {
-    *out_version = high_version;
-    return clnt;
-  }
-  while (low_version < high_version) {
-    if (rpc_stat != RPC_PROGVERSMISMATCH)
-      break;
-    clnt_geterr(clnt, &rpcerr);
-    lo = rpcerr.re_vers.low;
-    hi = rpcerr.re_vers.high;
-    if (hi < high_version)
-      high_version = hi;
-    else
-      high_version--;
-    if (lo > low_version)
-      low_version = lo;
-    if (low_version > high_version)
-      goto out;
+  timeout.tv_sec = 6;
+  /* Ping the portmapper on a remote system by calling the nullproc */
+  clnt_stat = clnt_call(client,
+			PMAPPROC_NULL,
+			(XDRPROC_T_TYPE) xdr_void,
+			NULL,
+			(XDRPROC_T_TYPE) xdr_void,
+			NULL,
+			timeout);
+  clnt_destroy(client);
+  close(socket);
+  sin->sin_port = 0;
 
-    CLNT_CONTROL(clnt, CLSET_VERS, (char *)&high_version);
-    rpc_stat = clnt_call(clnt,
-			 NULLPROC,
-			 (XDRPROC_T_TYPE) xdr_void,
-			 NULL,
-			 (XDRPROC_T_TYPE) xdr_void,
-			 NULL,
-			 tv);
-    if (rpc_stat == RPC_SUCCESS) {
-      *out_version = high_version;
-      return clnt;
-    }
+  if (clnt_stat == RPC_TIMEDOUT) {
+    plog(XLOG_ERROR,
+	 "check_pmap_up: failed to contact portmapper on host \"%s\": %s",
+	 host, clnt_sperrno(clnt_stat));
+    return 0;
   }
-  clnt_geterr(clnt, &rpcerr);
-
-out:
-  rpc_createerr.cf_stat = rpc_stat;
-  rpc_createerr.cf_error = rpcerr;
-  clnt_destroy(clnt);
-  return NULL;
+  return 1;
 }
 
 
@@ -637,7 +746,8 @@ u_long
 get_nfs_version(char *host, struct sockaddr_in *sin, u_long nfs_version, const char *proto)
 {
   CLIENT *clnt = NULL;
-  u_long versout;
+  rpcvers_t versout;
+  struct timeval tv;
 
   /*
    * If not set or set wrong, then try from NFS_VERS_MAX on down. If
@@ -647,7 +757,6 @@ get_nfs_version(char *host, struct sockaddr_in *sin, u_long nfs_version, const c
     nfs_version = NFS_VERS_MAX;
   }
 
-#ifdef DEBUG
   if (nfs_version == NFS_VERSION) {
     dlog("get_nfs_version trying NFS(%d,%s) for %s",
 	 (int) nfs_version, proto, host);
@@ -655,10 +764,17 @@ get_nfs_version(char *host, struct sockaddr_in *sin, u_long nfs_version, const c
     dlog("get_nfs_version trying NFS(%d-%d,%s) for %s",
 	 (int) NFS_VERSION, (int) nfs_version, proto, host);
   }
-#endif /* DEBUG */
-  /* get the best NFS version, and timeout quickly if remote host is down */
-  clnt = amu_clnt_create_best_vers(host, NFS_PROGRAM, &versout,
-				   NFS_VERSION, nfs_version, proto);
+
+  /* 3 seconds is more than enough for a LAN */
+  memset(&tv, 0, sizeof(tv));
+  tv.tv_sec = 3;
+  tv.tv_usec = 0;
+
+#ifdef HAVE_CLNT_CREATE_VERS_TIMED
+  clnt = clnt_create_vers_timed(host, NFS_PROGRAM, &versout, NFS_VERSION, nfs_version, proto, &tv);
+#else /* not HAVE_CLNT_CREATE_VERS_TIMED */
+  clnt = clnt_create_vers(host, NFS_PROGRAM, &versout, NFS_VERSION, nfs_version, proto);
+#endif	/* not HAVE_CLNT_CREATE_VERS_TIMED */
 
   if (clnt == NULL) {
     if (nfs_version == NFS_VERSION)
@@ -673,3 +789,136 @@ get_nfs_version(char *host, struct sockaddr_in *sin, u_long nfs_version, const c
 
   return versout;
 }
+
+
+#if defined(HAVE_FS_AUTOFS) && defined(AUTOFS_PROG)
+/*
+ * find the IP address that can be used to connect autofs service to.
+ */
+static int
+get_autofs_address(struct netconfig *ncp, struct t_bind *tbp)
+{
+  int ret;
+  struct nd_addrlist *addrs = (struct nd_addrlist *) NULL;
+  struct nd_hostserv service;
+
+  service.h_host = HOST_SELF_CONNECT;
+  service.h_serv = "autofs";
+
+  ret = netdir_getbyname(ncp, &service, &addrs);
+
+  if (ret) {
+    plog(XLOG_FATAL, "get_autofs_address: cannot get local host address: %s", netdir_sperror());
+    goto out;
+  }
+
+  /*
+   * XXX: there may be more more than one address for this local
+   * host.  Maybe something can be done with those.
+   */
+  tbp->addr.len = addrs->n_addrs->len;
+  tbp->addr.maxlen = addrs->n_addrs->len;
+  memcpy(tbp->addr.buf, addrs->n_addrs->buf, addrs->n_addrs->len);
+ /*
+  * qlen should not be zero for TCP connections.  It's not clear what it
+  * should be for UDP connections, but setting it to something like 64 seems
+  * to be the safe value that works.
+  */
+  tbp->qlen = 64;
+
+  /* all OK */
+  netdir_free((voidp) addrs, ND_ADDRLIST);
+
+out:
+  return ret;
+}
+
+
+/*
+ * Register the autofs service for amd
+ */
+int
+register_autofs_service(char *autofs_conftype, void (*autofs_dispatch)())
+{
+  struct t_bind *tbp = 0;
+  struct netconfig *autofs_ncp;
+  SVCXPRT *autofs_xprt = NULL;
+  int fd = -1, err = 1;		/* assume failed */
+
+  plog(XLOG_INFO, "registering autofs service: %s", autofs_conftype);
+  autofs_ncp = getnetconfigent(autofs_conftype);
+  if (autofs_ncp == NULL) {
+    plog(XLOG_ERROR, "register_autofs_service: cannot getnetconfigent for %s", autofs_conftype);
+    goto out;
+  }
+
+  fd = t_open(autofs_ncp->nc_device, O_RDWR, NULL);
+  if (fd < 0) {
+    plog(XLOG_ERROR, "register_autofs_service: t_open failed (%s)",
+	 t_errlist[t_errno]);
+    goto out;
+  }
+
+  tbp = (struct t_bind *) t_alloc(fd, T_BIND, T_ADDR);
+  if (!tbp) {
+    plog(XLOG_ERROR, "register_autofs_service: t_alloc failed");
+    goto out;
+  }
+
+  if (get_autofs_address(autofs_ncp, tbp) != 0) {
+    plog(XLOG_ERROR, "register_autofs_service: get_autofs_address failed");
+    goto out;
+  }
+
+  autofs_xprt = svc_tli_create(fd, autofs_ncp, tbp, 0, 0);
+  if (autofs_xprt == NULL) {
+    plog(XLOG_ERROR, "cannot create autofs tli service for amd");
+    goto out;
+  }
+
+  rpcb_unset(AUTOFS_PROG, AUTOFS_VERS, autofs_ncp);
+  if (svc_reg(autofs_xprt, AUTOFS_PROG, AUTOFS_VERS, autofs_dispatch, autofs_ncp) == FALSE) {
+    plog(XLOG_ERROR, "could not register amd AUTOFS service");
+    goto out;
+  }
+  err = 0;
+  goto really_out;
+
+out:
+  if (autofs_ncp)
+    freenetconfigent(autofs_ncp);
+  if (autofs_xprt)
+    SVC_DESTROY(autofs_xprt);
+  else {
+    if (fd > 0)
+      t_close(fd);
+  }
+
+really_out:
+  if (tbp)
+    t_free((char *) tbp, T_BIND);
+
+  dlog("register_autofs_service: returning %d\n", err);
+  return err;
+}
+
+
+int
+unregister_autofs_service(char *autofs_conftype)
+{
+  struct netconfig *autofs_ncp;
+  int err = 1;
+
+  plog(XLOG_INFO, "unregistering autofs service listener: %s", autofs_conftype);
+
+  autofs_ncp = getnetconfigent(autofs_conftype);
+  if (autofs_ncp == NULL) {
+    plog(XLOG_ERROR, "destroy_autofs_service: cannot getnetconfigent for %s", autofs_conftype);
+    goto out;
+  }
+
+out:
+  rpcb_unset(AUTOFS_PROG, AUTOFS_VERS, autofs_ncp);
+  return err;
+}
+#endif /* HAVE_FS_AUTOFS && AUTOFS_PROG */
