@@ -36,9 +36,8 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- *      %W% (Berkeley) %G%
  *
- * $Id: homedir.c,v 1.5.2.13 2005/01/03 20:56:14 ezk Exp $
+ * File: am-utils/hlfsd/homedir.c
  *
  * HLFSD was written at Columbia University Computer Science Department, by
  * Erez Zadok <ezk@cs.columbia.edu> and Alexander Dupuy <dupuy@cs.columbia.edu>
@@ -135,41 +134,18 @@ homedir(int userid, int groupid)
     }
   }
 
-#ifdef DEBUG
   /*
-   * only run this forking code if asked for -D fork
-   * or if did not ask for -D nofork
+   * only run this forking code if did not ask for -D fork
    */
-  amuDebug(D_FORK) {
-#endif /* DEBUG */
+  if (!amuDebug(D_FORK)) {
     /* fork child to process request if none in progress */
-    if (found->child && kill(found->child, 0) < 0)
+    if (found->child && kill(found->child, 0))
       found->child = 0;
 
     if (found->child)
       delay(found, 5);		/* wait a bit if in progress */
-
-#if defined(DEBUG) && defined(HAVE_WAITPID)
-    if (found->child) {
-      /* perhaps it's a child we lost count of? let's wait on it */
-      int status, child;
-      if ((child = waitpid((pid_t) found->child, &status, WNOHANG)) > 0) {
-	plog(XLOG_ERROR, "found lost child %d", child);
-	found->child = 0;
-	if (WIFEXITED(status))
-	  found->last_status = WEXITSTATUS(status);
-	else if (WIFSIGNALED(status))
-	  found->last_status = -WTERMSIG(status);
-	else {
-	  plog(XLOG_ERROR, "unknown child exit status (%d) ???", status);
-	  found->last_status = 255;
-	}
-      }
-    }
-#endif /* DEBUG && HAVE_WAITPID */
-
-    if (found->child) {
-      found->last_status = 1;	/* better safe than sorry - maybe */
+    if (found->child) {		/* better safe than sorry - maybe */
+      found->last_status = 1;
       return alt_spooldir;
     }
     if ((found->child = fork()) < 0) {
@@ -177,21 +153,17 @@ homedir(int userid, int groupid)
       return alt_spooldir;
     }
     if (found->child) {		/* PARENT */
-#ifdef DEBUG
       if (lastchild)
-	plog(XLOG_INFO, "cache spill uid = %ld, pid = %ld, home = %s",
+	dlog("cache spill uid = %ld, pid = %ld, home = %s",
 	     (long) lastchild->uid, (long) lastchild->child,
 	     lastchild->home);
-#endif /* DEBUG */
       lastchild = found;
       return (char *) NULL;	/* return NULL to parent, so it can continue */
     }
-#ifdef DEBUG
-  }				/* end of Debug(D_FORK) */
-#endif /* DEBUG */
+  }
 
   /*
-   * CHILD: (or parent if -D nofork)
+   * CHILD: (or parent if -D fork)
    *
    * Check and create dir if needed.
    * Check disk space and/or quotas too.
@@ -297,10 +269,8 @@ delay(uid2home_t *found, int secs)
 {
   struct timeval tv;
 
-#ifdef DEBUG
   if (found)
     dlog("delaying on child %ld for %d seconds", (long) found->child, secs);
-#endif /* DEBUG */
 
   tv.tv_usec = 0;
 
@@ -323,55 +293,30 @@ interlock(int signum)
   int child;
   uid2home_t *lostchild;
   int status;
-  int max_errors = 10;				/* avoid infinite loops */
 
 #ifdef HAVE_WAITPID
-  while ((child = waitpid((pid_t) -1, &status, WNOHANG)) != 0) {
+  while ((child = waitpid((pid_t) -1, &status, WNOHANG)) > 0) {
 #else /* not HAVE_WAITPID */
-  while ((child = wait3(&status, WNOHANG, (struct rusage *) 0)) != 0) {
+  while ((child = wait3(&status, WNOHANG, (struct rusage *) 0)) > 0) {
 #endif /* not HAVE_WAITPID */
 
-    if (child < 0) {
-      plog(XLOG_WARNING, "waitpid/wait3: %m");
-      if (--max_errors > 0)
-	continue;
-      else
-	break;
-    }
-
     /* high chances this was the last child forked */
     if (lastchild && lastchild->child == child) {
       lastchild->child = 0;
 
       if (WIFEXITED(status))
 	lastchild->last_status = WEXITSTATUS(status);
-      else if (WIFSIGNALED(status))
-	lastchild->last_status = -WTERMSIG(status);
-      else {
-	plog(XLOG_ERROR, "unknown child exit status (%d) ???", status);
-	lastchild->last_status = 255;
-      }
       lastchild = (uid2home_t *) NULL;
     } else {
       /* and if not, we have to search for it... */
-      int found = 0;
       for (lostchild = pwtab; lostchild < &pwtab[cur_pwtab_num]; lostchild++) {
 	if (lostchild->child == child) {
-	  lostchild->child = 0;
 	  if (WIFEXITED(status))
 	    lostchild->last_status = WEXITSTATUS(status);
-	  else if (WIFSIGNALED(status))
-	    lostchild->last_status = -WTERMSIG(status);
-	  else {
-	    plog(XLOG_ERROR, "unknown child exit status (%d) ???", status);
-	    lostchild->last_status = 255;
-	  }
-	  found = 1;
+	  lostchild->child = 0;
 	  break;
 	}
       }
-      if (!found)
-	plog(XLOG_ERROR, "no record of child %d found???", child);
     }
   }
 }
@@ -464,6 +409,7 @@ mailbox(int uid, char *username)
 
 static int
 plt_compare_fxn(const voidp x, const voidp y)
+
 {
   uid2home_t *i = (uid2home_t *) x;
   uid2home_t *j = (uid2home_t *) y;
@@ -726,10 +672,8 @@ table_add(u_int u, const char *h, const char *n)
   /* do NOT add duplicate entries (this is an O(N^2) algorithm... */
   for (i=0; i<cur_pwtab_num; ++i)
     if (u == pwtab[i].uid  &&  u != 0 ) {
-#ifdef DEBUG
       dlog("ignoring duplicate home %s for uid %d (already %s)",
 	   h, u, pwtab[i].home);
-#endif /* DEBUG */
       return;
     }
 
