@@ -36,9 +36,8 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- *      %W% (Berkeley) %G%
  *
- * $Id: stubs.c,v 1.5.2.8 2005/01/03 20:56:14 ezk Exp $
+ * File: am-utils/hlfsd/stubs.c
  *
  * HLFSD was written at Columbia University Computer Science Department, by
  * Erez Zadok <ezk@cs.columbia.edu> and Alexander Dupuy <dupuy@cs.columbia.edu>
@@ -127,8 +126,18 @@ nfsproc_getattr_2_svc(am_nfs_fh *argp, struct svc_req *rqstp)
     res.ns_status = NFSERR_STALE;
     return &res;
   }
-
   if (eq_fh(argp, &root)) {
+#if 0
+    /*
+     * XXX: increment mtime of parent directory, causes NFS clients to
+     * invalidate their cache for that directory.
+     * Some NFS clients may need this code.
+     */
+    if (uid != rootfattr.na_uid) {
+      rootfattr.na_mtime.nt_seconds++;
+      rootfattr.na_uid = uid;
+    }
+#endif
     res.ns_status = NFS_OK;
     res.ns_u.ns_attr_u = rootfattr;
   } else if (eq_fh(argp, &slink)) {
@@ -161,10 +170,8 @@ nfsproc_getattr_2_svc(am_nfs_fh *argp, struct svc_req *rqstp)
 	res.ns_status = NFS_OK;
 	un_fattr.na_fileid = uid;
 	res.ns_u.ns_attr_u = un_fattr;
-#ifdef DEBUG
 	dlog("nfs_getattr: successful search for uid=%ld, gid=%ld",
 	     (long) uid, (long) gid);
-#endif /* DEBUG */
       } else {			/* not found */
 	res.ns_status = NFSERR_STALE;
       }
@@ -219,12 +226,22 @@ nfsproc_lookup_2_svc(nfsdiropargs *argp, struct svc_req *rqstp)
     res.dr_status = NFSERR_NOENT;
     return &res;
   }
-
   if (eq_fh(&argp->da_fhandle, &root)) {
     if (argp->da_name[0] == '.' &&
 	(argp->da_name[1] == '\0' ||
 	 (argp->da_name[1] == '.' &&
 	  argp->da_name[2] == '\0'))) {
+#if 0
+    /*
+     * XXX: increment mtime of parent directory, causes NFS clients to
+     * invalidate their cache for that directory.
+     * Some NFS clients may need this code.
+     */
+      if (uid != rootfattr.na_uid) {
+	rootfattr.na_mtime.nt_seconds++;
+	rootfattr.na_uid = uid;
+      }
+#endif
       res.dr_u.dr_drok_u.drok_fhandle = root;
       res.dr_u.dr_drok_u.drok_attributes = rootfattr;
       res.dr_status = NFS_OK;
@@ -267,15 +284,13 @@ nfsproc_lookup_2_svc(nfsdiropargs *argp, struct svc_req *rqstp)
       res.dr_u.dr_drok_u.drok_attributes = un_fattr;
       memset((char *) &un_fhandle, 0, sizeof(am_nfs_fh));
       *(u_int *) un_fhandle.fh_data = (u_int) untab[idx].uid;
-      strncpy((char *) &un_fhandle.fh_data[sizeof(int)],
-	      untab[idx].username,
-	      sizeof(am_nfs_fh) - sizeof(int));
+      xstrlcpy((char *) &un_fhandle.fh_data[sizeof(int)],
+	       untab[idx].username,
+	       sizeof(am_nfs_fh) - sizeof(int));
       res.dr_u.dr_drok_u.drok_fhandle = un_fhandle;
       res.dr_status = NFS_OK;
-#ifdef DEBUG
       dlog("nfs_lookup: successful lookup for uid=%ld, gid=%ld: username=%s",
 	   (long) uid, (long) gid, untab[idx].username);
-#endif /* DEBUG */
       return &res;
     }
   } /* end of "if (eq_fh(argp->dir.data, root.data)) {" */
@@ -340,7 +355,7 @@ nfsproc_readlink_2_svc(am_nfs_fh *argp, struct svc_req *rqstp)
     last_uid = userid;
   }
 
-  /* I don't think will pass this if -D nofork */
+  /* I don't think it will pass this if -D fork */
   if (serverpid == getpid())
     return &res;
 
@@ -359,15 +374,13 @@ nfsproc_readlink_2_svc(am_nfs_fh *argp, struct svc_req *rqstp)
   else
     retval = 0;
 
-#ifdef DEBUG
   /*
-   * If asked for -D nofork, then must return the value,
+   * If asked for -D fork, then must return the value,
    * NOT exit, or else the main hlfsd server exits.
-   * Bug where is that status information being collected?
+   * Bug: where is that status information being collected?
    */
-  amuDebugNo(D_FORK)
+  if (amuDebug(D_FORK))
     return &res;
-#endif /* DEBUG */
 
   exit(retval);
 }
